/*
 * Copyright (C) 2005-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file net.c Networking-related helper functions. */

/*
 * Since glibc 2.8, the _GNU_SOURCE feature test macro must be defined in order
 * to obtain the definition of the ucred structure.
 */
#define _GNU_SOURCE

#include <netdb.h>

/* At least NetBSD needs these. */
#ifndef AI_V4MAPPED
#define AI_V4MAPPED 0
#endif
#ifndef AI_ALL
#define AI_ALL 0
#endif
#ifndef AI_ADDRCONFIG
#define AI_ADDRCONFIG 0
#endif

#include <dirent.h>
#include <regex.h>

#include "para.h"
#include "error.h"
#include "net.h"
#include "string.h"
#include "fd.h"


/** Information about one encrypted connection. */
struct crypt_data {
	/** Function used to decrypt received data. */
	crypt_function *recv;
	/** Function used to encrypt data to be sent. */
	crypt_function *send;
	/**
	 * Context-dependent data (crypt keys), passed verbatim to the above
	 * crypt functions.
	 */
	void *private_data;
};
/** Array holding per fd crypt data. */
static struct crypt_data *crypt_data_array;
/** Current size of the crypt data array. */
static unsigned cda_size = 0;

/**
 * Activate encryption for one file descriptor.
 *
 * \param fd The file descriptor.
 * \param recv_f The function used for decrypting received data.
 * \param send_f The function used for encrypting before sending.
 * \param private_data User data supplied by the caller.
 */
void enable_crypt(int fd, crypt_function *recv_f, crypt_function *send_f,
	void *private_data)
{
	if (fd + 1 > cda_size) {
		crypt_data_array = para_realloc(crypt_data_array,
			(fd + 1) * sizeof(struct crypt_data));
		memset(crypt_data_array + cda_size, 0,
			(fd + 1 - cda_size) * sizeof(struct crypt_data));
		cda_size = fd + 1;
	}
	crypt_data_array[fd].recv = recv_f;
	crypt_data_array[fd].send = send_f;
	crypt_data_array[fd].private_data = private_data;
	PARA_INFO_LOG("rc4 encryption activated for fd %d\n", fd);
}

/**
 * Deactivate encryption for a given fd.
 *
 * \param fd The file descriptor.
 *
 * This must be called if and only if \p fd was activated via enable_crypt().
 */
void disable_crypt(int fd)
{
	if (cda_size < fd + 1)
		return;
	crypt_data_array[fd].recv = NULL;
	crypt_data_array[fd].send = NULL;
	crypt_data_array[fd].private_data = NULL;
}

/**
 * Match string as a candidate IPv4 address.
 *
 * \param address The string to match.
 * \return True if \a address has "dot-quad" format.
 */
static bool is_v4_dot_quad(const char *address)
{
	bool result;
	regex_t r;

	assert(!regcomp(&r, "^([0-9]+\\.){3}[0-9]+$", REG_EXTENDED|REG_NOSUB));
	result = regexec(&r, address, 0, NULL, 0) == 0;
	regfree(&r);
	return result;
}

/**
 * Perform basic syntax checking on the host-part of an URL:
 *
 * - Since ':' is invalid in IPv4 addresses and DNS names, the
 *   presence of ':' causes interpretation as IPv6 address;
 * - next the first-match-wins algorithm from RFC 3986 is applied;
 * - else the string is considered as DNS name, to be resolved later.
 *
 * \param host The host string to check.
 * \return True if \a host passes the syntax checks.
 *
 * \sa RFC 3986, 3.2.2; RFC 1123, 2.1; RFC 1034, 3.5
 */
static bool host_string_ok(const char *host)
{
	if (host == NULL || *host == '\0')
		return false;
	if (strchr(host, ':') != NULL)
		return is_valid_ipv6_address(host);
	if (is_v4_dot_quad(host))
		return is_valid_ipv4_address(host);
	return true;
}

/**
 * Parse and validate URL string.
 *
 * The URL syntax is loosely based on RFC 3986, supporting one of
 * - "["host"]"[:port] for native IPv6 addresses and
 * - host[:port] for IPv4 hostnames and DNS names.
 *
 * Native IPv6 addresses must be enclosed in square brackets, since
 * otherwise there is an ambiguity with the port separator `:'.
 * The 'port' part is always considered to be a number; if absent,
 * it is set to -1, to indicate that a default port is to be used.
 *
 * The following are valid examples:
 * - 10.10.1.1
 * - 10.10.1.2:8000
 * - localhost
 * - localhost:8001
 * - [::1]:8000
 * - [badc0de::1]
 *
 * \param url	  The URL string to take apart.
 * \param host	  To return the copied host part of \a url.
 * \param hostlen The maximum length of \a host.
 * \param port	  To return the port number (if any) of \a url.
 *
 * \return Pointer to \a host, or NULL if failed.
 * If NULL is returned, \a host and \a portnum are undefined. If no
 * port number was present in \a url, \a portnum is set to -1.
 *
 * \sa RFC 3986, 3.2.2/3.2.3
 */
char *parse_url(const char *url,
		char	*host, ssize_t hostlen,
		int32_t	*port)
{
	const char *o = url;
	char *c = host, *end = c + (hostlen - 1);

	*port = -1;

	if (o == NULL || hostlen < 1)
		goto failed;

	if (*o == '[') {
		for (++o; (*c = *o == ']' ? '\0' : *o); c++, o++)
			if (c == end)
				goto failed;

		if (*o++ != ']' || (*o != '\0' && *o != ':'))
			goto failed;
	} else {
		for (; (*c = *o == ':'? '\0' : *o); c++, o++)
			if (c == end)
				goto failed;
	}

	if (*o == ':')
		if (para_atoi32(++o, port) < 0 ||
		    *port < 0 || *port > 0xffff)
			goto failed;

	if (host_string_ok(host))
		return host;
failed:
	*host = '\0';
	return NULL;
}

/**
 * Determine the socket type for a given layer-4 protocol.
 *
 * \param l4type The symbolic name of the transport-layer protocol.
 *
 * \sa ip(7), socket(2)
 */
static inline int sock_type(const unsigned l4type)
{
	switch (l4type) {
	case IPPROTO_UDP:	return SOCK_DGRAM;
	case IPPROTO_TCP:	return SOCK_STREAM;
	case IPPROTO_DCCP:	return SOCK_DCCP;
	}
	return -1;		/* not supported here */
}

/**
 * Pretty-print transport-layer name.
 */
static const char *layer4_name(const unsigned l4type)
{
	switch (l4type) {
	case IPPROTO_UDP:       return "UDP";
	case IPPROTO_TCP:       return "TCP";
	case IPPROTO_DCCP:      return "DCCP";
	}
	return "UNKNOWN PROTOCOL";
}

/**
 * Resolve IPv4/IPv6 address and create a ready-to-use active or passive socket.
 *
 * \param l3type The layer-3 type (\p AF_INET, \p AF_INET6, \p AF_UNSPEC).
 * \param l4type The layer-4 type (\p IPPROTO_xxx).
 * \param passive Whether this is a passive (1) or active (0) socket.
 * \param host Remote or local hostname or IPv/6 address string.
 * \param port_number Decimal port number.
 *
 * This creates a ready-made IPv4/v6 socket structure after looking up the
 * necessary parameters. The interpretation of \a host depends on the value of
 * \a passive:
 *	- on a passive socket host is interpreted as an interface IPv4/6 address
 *	  (can be left NULL);
 *	- on an active socket, \a host is the peer DNS name or IPv4/6 address
 *	  to connect to;
 * 	- \a port_number is in either case the numeric port number (not service
 * 	  string).
 *
 * Furthermore, bind(2) is called on passive sockets, and connect(2) on active
 * sockets. The algorithm tries all possible address combinations until it
 * succeeds.
 *
 * \return This function returns 1 on success and \a -E_ADDRESS_LOOKUP when no
 * matching connection could be set up (with details in the error log).
 *
 *  \sa ipv6(7), getaddrinfo(3), bind(2), connect(2).
 */
int makesock(unsigned l3type, unsigned l4type, int passive,
		const char *host, unsigned short port_number)
{
	struct addrinfo *local = NULL, *src,
			*remote = NULL, *dst, hints;
	int		rc, on = 1, sockfd = -1,
			socktype = sock_type(l4type);
	char port[6]; /* port number has at most 5 digits */

	sprintf(port, "%u", port_number);
	/* Set up address hint structure */
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = l3type;
	hints.ai_socktype = socktype;
	/* 
	 * getaddrinfo does not support SOCK_DCCP, so for the sake of lookup
	 * (and only then) pretend to be UDP.
	 */
	if (l4type == IPPROTO_DCCP)
		hints.ai_socktype = SOCK_DGRAM;

	/* only use addresses available on the host */
	hints.ai_flags = AI_ADDRCONFIG;
	if (l3type == AF_INET6)
		/* use v4-mapped-v6 if no v6 addresses found */
		hints.ai_flags |= AI_V4MAPPED | AI_ALL;

	if (passive && host == NULL)
		hints.ai_flags |= AI_PASSIVE;

	/* Obtain local/remote address information */
	if ((rc = getaddrinfo(host, port, &hints, passive ? &local : &remote))) {
		PARA_ERROR_LOG("can not resolve %s address %s#%s: %s.\n",
				layer4_name(l4type),
				host? host : (passive? "[loopback]" : "[localhost]"),
				port, gai_strerror(rc));
		return -E_ADDRESS_LOOKUP;
	}

	/* Iterate over all src/dst combination, exhausting dst first */
	for (src = local, dst = remote; src != NULL || dst != NULL; /* no op */ ) {
		if (src && dst && src->ai_family == AF_INET
				&& dst->ai_family == AF_INET6)
			goto get_next_dst; /* v4 -> v6 is not possible */

		sockfd = socket(src ? src->ai_family : dst->ai_family,
			socktype, l4type);
		if (sockfd < 0)
			goto get_next_dst;

		/*
		 * Set those options that need to be set before establishing
		 * the connection. Reuse the address on passive (listening)
		 * sockets to avoid failure on restart.
		 */
		if (passive && setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
				&on, sizeof(on)) == -1) {
			PARA_ERROR_LOG("can not set SO_REUSEADDR: %s\n",
				strerror(errno));
			return -ERRNO_TO_PARA_ERROR(errno);
		}

		if (src) {
			if (bind(sockfd, src->ai_addr, src->ai_addrlen) < 0) {
				close(sockfd);
				goto get_next_src;
			}
			if (!dst) /* bind-only completed successfully */
				break;
		}

		if (dst && connect(sockfd, dst->ai_addr, dst->ai_addrlen) == 0)
			break; /* connection completed successfully */
		close(sockfd);
get_next_dst:
		if (dst && (dst = dst->ai_next))
			continue;
get_next_src:
		if (src && (src = src->ai_next)) /* restart inner loop */
			dst = remote;
	}
	if (local)
		freeaddrinfo(local);
	if (remote)
		freeaddrinfo(remote);

	if (src == NULL && dst == NULL) {
		PARA_ERROR_LOG("can not create %s socket %s#%s.\n",
			layer4_name(l4type), host? host : (passive?
			"[loopback]" : "[localhost]"), port);
		return -ERRNO_TO_PARA_ERROR(errno);
	}
	return sockfd;
}

/**
 * Create a passive / listening socket.
 *
 * \param l3type The network-layer type (\p AF_xxx).
 * \param l4type The transport-layer type (\p IPPROTO_xxx).
 * \param port The decimal port number to listen on.
 *
 * \return Positive integer (socket descriptor) on success, negative value
 * otherwise.
 *
 * \sa makesock(), ip(7), ipv6(7), bind(2), listen(2).
 */
int para_listen(unsigned l3type, unsigned l4type, unsigned short port)
{
	int ret, fd = makesock(l3type, l4type, 1, NULL, port);

	if (fd > 0) {
		ret = listen(fd, BACKLOG);
		if (ret < 0) {
			close(fd);
			return -ERRNO_TO_PARA_ERROR(errno);
		}
		PARA_INFO_LOG("listening on %s port %u, fd %d\n",
			layer4_name(l4type), port, fd);
	}
	return fd;
}

/**
 * Print numeric host and port number (beware - uses static char).
 *
 * \param sa The IPv4/IPv6 socket address to use.
 * \param len The length of \p sa.
 *
 * \sa getnameinfo(3).
 */
static char *host_and_port(struct sockaddr *sa, socklen_t len)
{
	static char output[NI_MAXHOST + NI_MAXSERV + 2];
	char hbuf[NI_MAXHOST], sbuf[NI_MAXSERV];
	int ret;

	ret = getnameinfo(sa, len, hbuf, sizeof(hbuf), sbuf, sizeof(sbuf),
		NI_NUMERICHOST | NI_NUMERICSERV);
	if (ret) {
		PARA_WARNING_LOG("hostname lookup error (%s).\n",
			gai_strerror(ret));
		sprintf(output, "(unknown)");
	} else
		sprintf(output, "%s#%s", hbuf, sbuf);
	return output;
}

/**
 * Look up the local or remote side of a connected socket structure.
 *
 * \param fd The socket descriptor of the connected socket.
 * \param getname Either \p getsockname() for local, or \p getpeername() for
 * remote side.
 *
 * \return A static character string identifying hostname and port of the
 * chosen side.
 *
 * \sa getsockname(2), getpeername(2).
 */
static char *__get_sock_name(int fd, int (*getname)(int, struct sockaddr*,
		socklen_t *))
{
	struct sockaddr_storage ss;
	socklen_t sslen = sizeof(ss);

	if (getname(fd, (struct sockaddr *)&ss, &sslen) < 0) {
		static char *dont_know = "(don't know)";
		PARA_ERROR_LOG("can not determine address from fd %d: %s\n",
			fd, strerror(errno));
		return dont_know;
	}
	return host_and_port((struct sockaddr *)&ss, sslen);
}

char *local_name(int sockfd)
{
	return __get_sock_name(sockfd, getsockname);
}

char *remote_name(int sockfd)
{
	return __get_sock_name(sockfd, getpeername);
}

/**
 * Extract IPv4 or IPv6-mapped-IPv4 address from sockaddr_storage.
 * \param ss Container of IPv4/6 address
 * \return Extracted IPv4 address (different from 0) or 0 if unsuccessful.
 *
 * \sa RFC 3493
 */
struct in_addr extract_v4_addr(const struct sockaddr_storage *ss)
{
	struct in_addr ia = {.s_addr = 0};

	if (ss->ss_family == AF_INET)
		 ia.s_addr = ((struct sockaddr_in *)ss)->sin_addr.s_addr;
	if (ss->ss_family == AF_INET6) {
		const struct in6_addr v6_addr = ((struct sockaddr_in6 *)ss)->sin6_addr;

		if (IN6_IS_ADDR_V4MAPPED(&v6_addr))
			memcpy(&ia.s_addr, &(v6_addr.s6_addr[12]), 4);
	}
	return ia;
}

/**
 * Encrypt and send a binary buffer.
 *
 * \param fd The file descriptor.
 * \param buf The buffer to be encrypted and sent.
 * \param len The length of \a buf.
 *
 * Check if encryption is available. If yes, encrypt the given buffer.  Send
 * out the buffer, encrypted or not, and try to resend the remaing part in case
 * of short writes.
 *
 * \return Standard.
 */
int send_bin_buffer(int fd, const char *buf, size_t len)
{
	int ret;
	crypt_function *cf = NULL;

	if (!len)
		PARA_CRIT_LOG("len == 0\n");
	if (fd + 1 <= cda_size)
		cf = crypt_data_array[fd].send;
	if (cf) {
		void *private = crypt_data_array[fd].private_data;
		/* RC4 may write more than len to the output buffer */
		unsigned char *outbuf = para_malloc(ROUND_UP(len, 8));
		(*cf)(len, (unsigned char *)buf, outbuf, private);
		ret = write_all(fd, (char *)outbuf, &len);
		free(outbuf);
	} else
		ret = write_all(fd, buf, &len);
	return ret;
}

/**
 * Encrypt and send null terminated buffer.
 *
 * \param fd The file descriptor.
 * \param buf The null-terminated buffer to be send.
 *
 * This is equivalent to send_bin_buffer(fd, buf, strlen(buf)).
 *
 * \return Standard.
 */
int send_buffer(int fd, const char *buf)
{
	return send_bin_buffer(fd, buf, strlen(buf));
}


/**
 * Send and encrypt a buffer given by a format string.
 *
 * \param fd The file descriptor.
 * \param fmt A format string.
 *
 * \return Standard.
 */
__printf_2_3 int send_va_buffer(int fd, const char *fmt, ...)
{
	char *msg;
	int ret;

	PARA_VSPRINTF(fmt, msg);
	ret = send_buffer(fd, msg);
	free(msg);
	return ret;
}

/**
 * Receive and decrypt.
 *
 * \param fd The file descriptor.
 * \param buf The buffer to write the decrypted data to.
 * \param size The size of \a buf.
 *
 * Receive at most \a size bytes from file descriptor \a fd. If encryption is
 * available, decrypt the received buffer.
 *
 * \return The number of bytes received on success, negative on errors.
 *
 * \sa recv(2)
 */
__must_check int recv_bin_buffer(int fd, char *buf, size_t size)
{
	ssize_t n;
	crypt_function *cf = NULL;

	if (fd + 1 <= cda_size)
		cf = crypt_data_array[fd].recv;
	if (cf) {
		unsigned char *tmp = para_malloc(size);
		void *private = crypt_data_array[fd].private_data;
		n = recv(fd, tmp, size, 0);
		if (n > 0) {
			size_t numbytes = n;
			unsigned char *b = (unsigned char *)buf;
			(*cf)(numbytes, tmp, b, private);
		}
		free(tmp);
	} else
		n = recv(fd, buf, size, 0);
	if (n == -1)
		return -ERRNO_TO_PARA_ERROR(errno);
	return n;
}

/**
 * Receive, decrypt and write terminating NULL byte.
 *
 * \param fd The file descriptor.
 * \param buf The buffer to write the decrypted data to.
 * \param size The size of \a buf.
 *
 * Read and decrypt at most \a size - 1 bytes from file descriptor \a fd and
 * write a NULL byte at the end of the received data.
 *
 * \return The return value of the underlying call to \a recv_bin_buffer().
 *
 * \sa recv_bin_buffer()
 */
int recv_buffer(int fd, char *buf, size_t size)
{
	int n;

	assert(size);
	n = recv_bin_buffer(fd, buf, size - 1);
	if (n >= 0)
		buf[n] = '\0';
	else
		*buf = '\0';
	return n;
}

/**
 * Wrapper around the accept system call.
 *
 * \param fd The listening socket.
 * \param addr Structure which is filled in with the address of the peer socket.
 * \param size Should contain the size of the structure pointed to by \a addr.
 *
 * Accept incoming connections on \a addr. Retry if interrupted.
 *
 * \return The new file descriptor on success, negative on errors.
 *
 * \sa accept(2).
 */
int para_accept(int fd, void *addr, socklen_t size)
{
	int new_fd;

	do
		new_fd = accept(fd, (struct sockaddr *) addr, &size);
	while (new_fd < 0 && errno == EINTR);
	return new_fd < 0? -ERRNO_TO_PARA_ERROR(errno) : new_fd;
}

/**
 * Prepare a structure for \p AF_UNIX socket addresses.
 *
 * \param u Pointer to the struct to be prepared.
 * \param name The socket pathname.
 *
 * This just copies \a name to the sun_path component of \a u.
 *
 * \return Positive on success, \p -E_NAME_TOO_LONG if \a name is longer
 * than \p UNIX_PATH_MAX.
 */
static int init_unix_addr(struct sockaddr_un *u, const char *name)
{
	if (strlen(name) >= UNIX_PATH_MAX)
		return -E_NAME_TOO_LONG;
	memset(u->sun_path, 0, UNIX_PATH_MAX);
	u->sun_family = PF_UNIX;
	strcpy(u->sun_path, name);
	return 1;
}

/**
 * Prepare, create, and bind a socket for local communication.
 *
 * \param name The socket pathname.
 * \param unix_addr Pointer to the \p AF_UNIX socket structure.
 * \param mode The desired mode of the socket.
 *
 * This function creates a local socket for sequenced, reliable,
 * two-way, connection-based byte streams.
 *
 * \return The file descriptor, on success, negative on errors.
 *
 * \sa socket(2)
 * \sa bind(2)
 * \sa chmod(2)
 */
int create_local_socket(const char *name, struct sockaddr_un *unix_addr,
		mode_t mode)
{
	int fd, ret;

	ret = init_unix_addr(unix_addr, name);
	if (ret < 0)
		return ret;
	ret = socket(PF_UNIX, SOCK_STREAM, 0);
	if (ret < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	fd = ret;
	ret = bind(fd, (struct sockaddr *) unix_addr, UNIX_PATH_MAX);
	if (ret < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	ret = -E_CHMOD;
	if (chmod(name, mode) < 0)
		goto err;
	return fd;
err:
	close(fd);
	return ret;
}

/**
 * Prepare, create, and connect to a Unix domain socket for local communication.
 *
 * \param name The socket pathname.
 *
 * This function creates a local socket for sequenced, reliable, two-way,
 * connection-based byte streams.
 *
 * \return The file descriptor, on success, negative on errors.
 *
 * \sa create_local_socket(), unix(7), connect(2).
 */
int create_remote_socket(const char *name)
{
	struct sockaddr_un unix_addr;
	int fd, ret;

	ret = init_unix_addr(&unix_addr, name);
	if (ret < 0)
		return ret;
	fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (fd < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	if (connect(fd, (struct sockaddr *)&unix_addr, sizeof(unix_addr)) == -1) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	return fd;
err:
	close(fd);
	return ret;
}

#ifndef HAVE_UCRED
ssize_t send_cred_buffer(int sock, char *buf)
{
	return send_buffer(sock, buf);
}
int recv_cred_buffer(int fd, char *buf, size_t size)
{
	return recv_buffer(fd, buf, size) > 0? 1 : -E_RECVMSG;
}
#else /* HAVE_UCRED */
/**
 * Send \p NULL-terminated buffer and Unix credentials of the current process.
 *
 * \param sock The socket file descriptor.
 * \param buf The buffer to be sent.
 *
 * \return On success, this call returns the number of characters sent.  On
 * error, \p -E_SENDMSG is returned.
 *
 * \sa sendmsg(2), okir's Black Hats Manual.
 */
ssize_t send_cred_buffer(int sock, char *buf)
{
	char control[sizeof(struct cmsghdr) + sizeof(struct ucred)];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	static struct iovec iov;
	struct ucred c;
	int ret;

	/* Response data */
	iov.iov_base = buf;
	iov.iov_len  = strlen(buf);
	c.pid = getpid();
	c.uid = getuid();
	c.gid = getgid();
	/* compose the message */
	memset(&msg, 0, sizeof(msg));
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);
	/* attach the ucred struct */
	cmsg = CMSG_FIRSTHDR(&msg);
	cmsg->cmsg_level = SOL_SOCKET;
	cmsg->cmsg_type = SCM_CREDENTIALS;
	cmsg->cmsg_len = CMSG_LEN(sizeof(struct ucred));
	*(struct ucred *)CMSG_DATA(cmsg) = c;
	msg.msg_controllen = cmsg->cmsg_len;
	ret = sendmsg(sock, &msg, 0);
	if (ret  < 0)
		ret = -E_SENDMSG;
	return ret;
}

static void dispose_fds(int *fds, unsigned num)
{
	int i;

	for (i = 0; i < num; i++)
		close(fds[i]);
}

/**
 * Receive a buffer and the Unix credentials of the sending process.
 *
 * \param fd the socket file descriptor.
 * \param buf the buffer to store the message.
 * \param size the size of \a buffer.
 *
 * \return negative on errors, the user id on success.
 *
 * \sa recvmsg(2), okir's Black Hats Manual.
 */
int recv_cred_buffer(int fd, char *buf, size_t size)
{
	char control[255];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	struct iovec iov;
	int result = 0;
	int yes = 1;
	struct ucred cred;

	setsockopt(fd, SOL_SOCKET, SO_PASSCRED, &yes, sizeof(int));
	memset(&msg, 0, sizeof(msg));
	memset(buf, 0, size);
	iov.iov_base = buf;
	iov.iov_len = size;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);
	if (recvmsg(fd, &msg, 0) < 0)
		return -E_RECVMSG;
	result = -E_SCM_CREDENTIALS;
	cmsg = CMSG_FIRSTHDR(&msg);
	while (cmsg) {
		if (cmsg->cmsg_level == SOL_SOCKET && cmsg->cmsg_type
				== SCM_CREDENTIALS) {
			memcpy(&cred, CMSG_DATA(cmsg), sizeof(struct ucred));
			result = cred.uid;
		} else
			if (cmsg->cmsg_level == SOL_SOCKET
					&& cmsg->cmsg_type == SCM_RIGHTS) {
				dispose_fds((int *) CMSG_DATA(cmsg),
					(cmsg->cmsg_len - CMSG_LEN(0))
					/ sizeof(int));
			}
		cmsg = CMSG_NXTHDR(&msg, cmsg);
	}
	return result;
}
#endif /* HAVE_UCRED */

/**
 * Receive a buffer and check for a pattern.
 *
 * \param fd The file descriptor to receive from.
 * \param pattern The expected pattern.
 * \param bufsize The size of the internal buffer.
 *
 * \return Positive if \a pattern was received, negative otherwise.
 *
 * This function tries to receive at most \a bufsize bytes from file descriptor
 * \a fd. If at least \p strlen(\a pattern) bytes were received, the beginning
 * of the received buffer is compared with \a pattern, ignoring case.
 *
 * \sa recv_buffer(), \sa strncasecmp(3).
 */
int recv_pattern(int fd, const char *pattern, size_t bufsize)
{
	size_t len = strlen(pattern);
	char *buf = para_malloc(bufsize + 1);
	int ret = -E_RECV_PATTERN, n = recv_buffer(fd, buf, bufsize + 1);

	if (n < len)
		goto out;
	if (strncasecmp(buf, pattern, len))
		goto out;
	ret = 1;
out:
	if (ret < 0) {
		PARA_NOTICE_LOG("n = %d, did not receive pattern '%s'\n", n,
			pattern);
		if (n > 0)
			PARA_NOTICE_LOG("recvd: %s\n", buf);
	}
	free(buf);
	return ret;
}
